//===- Dialect.h - Enzyme dialect -------------------------------*- C++ -*-===//
//
// This file is licensed under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef ENZYME_DIALECT_H
#define ENZYME_DIALECT_H

#include "mlir/IR/Dialect.h"

#include "Dialect/EnzymeOpsDialect.h.inc"

#endif // ENZYME_DIALECT_H
