LIBRARY extern
EXPORTS
    LooksLikeAPrivateGlobal
    cdecl_fn_undecorated
    cdecl_fn_undecorated2
    cdecl_fn_noprefix
    cdecl_fn_decorated
    stdcall_fn_undecorated
    stdcall_fn_undecorated2
    stdcall_fn_noprefix@4
    fastcall_fn_undecorated
    fastcall_fn_undecorated2
    @fastcall_fn_decorated@4

    ;ld doesn't handle fully-decorated stdcall, or no-prefix fastcall
    _stdcall_fn_decorated@4=stdcall_fn_decorated@4
    fastcall_fn_noprefix@4=@fastcall_fn_noprefix@4

    ;Variables are never decorated
    extern_variable_undecorated
    extern_variable_noprefix
    extern_variable_decorated
